import asyncio
import sys
from typing import AsyncIterator


async def async_getch() -> AsyncIterator[bytes]:
    """An async iterator that yields characters read from the terminal in a non-blocking way."""

    is_windows = sys.platform == "win32"

    if is_windows:
        # Haven't actually tested this on Windows yet, hopefully it works
        import msvcrt

        getch = msvcrt.getch  # type: ignore

        def cleanup() -> None:
            pass

    else:
        import termios
        import tty

        # Save the current terminal settings
        old_settings = termios.tcgetattr(sys.stdin)

        # Set the terminal to cbreak mode
        tty.setcbreak(sys.stdin)

        getch = sys.stdin.buffer.read1  # type: ignore

        def cleanup() -> None:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)

    # on_char_ready will be called when stdin has a character;
    # it will put the character in a queue, and we'll later
    # yield from it
    char_queue: asyncio.Queue[bytes] = asyncio.Queue()

    def on_char_ready() -> None:
        bs: bytes = getch()
        char_queue.put_nowait(bs)

    # Add a reader for sys.stdin to call on_char_ready when appropriate
    loop = asyncio.get_running_loop()
    loop.add_reader(sys.stdin, on_char_ready)

    try:
        while True:
            yield await char_queue.get()
    finally:
        # Remove the reader for sys.stdin and restore the previous blocking + tty state
        loop.remove_reader(sys.stdin)
        cleanup()


__all__ = ("async_getch",)



from __future__ import annotations

import asyncio
import functools
import time
from collections.abc import AsyncIterator, Iterable, AsyncIterable

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

from typing import Callable, cast, Coroutine, TYPE_CHECKING, TypeVar

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)


_P = ParamSpec("_P")

_A = TypeVar("_A")
_B = TypeVar("_B")


def ensure_coroutine_function(
    fn: Callable[_P, _T_co] | Callable[_P, Coroutine[_A, _B, _T_co]],
    to_thread: bool = False,
) -> Callable[_P, Coroutine[_A, _B, _T_co]]:
    """Given a sync or async function, return an async function.

    Args:
        fn: The function to ensure is async.
        to_thread: Whether to run the function in a thread, if it is sync.

    Returns:
        An async function that runs the original function.
    """

    if asyncio.iscoroutinefunction(fn):
        return fn

    _fn_sync = cast(Callable[_P, _T_co], fn)
    if to_thread:

        @functools.wraps(_fn_sync)
        async def _async_fn(*__args: _P.args, **__kwargs: _P.kwargs) -> _T_co:
            return await asyncio.to_thread(_fn_sync, *__args, **__kwargs)

    else:

        @functools.wraps(_fn_sync)
        async def _async_fn(*__args: _P.args, **__kwargs: _P.kwargs) -> _T_co:
            return _fn_sync(*__args, **__kwargs)

    return _async_fn


async def iter_to_aiter(iterable: Iterable[_T], /, target_dt: float = 0.0005) -> AsyncIterator[_T]:
    """Convert an iterable to an async iterator, running the iterable in a
    thread.

        Args:
            iterable: The iterable to convert.
            target_dt: The target maximum time to possibly block the event loop
    for. Defaults to 0.0005.
                Note that this is not a hard limit, and the actual time spent
    blocking the event loop
                may be longer than this. See also sys.getswitchinterval().

        Yields:
            Items from the iterable.

        Examples:
            >>> async def demo_iter_to_thread():
            ...     async for item in iter_to_aiter(range(5)):
            ...         print(item)
            >>> asyncio.run(demo_iter_to_thread())
            0
            1
            2
            3
            4
    """
    iterator = iter(iterable)
    loop = asyncio.get_running_loop()
    result: list[_T] = []

    def run_iterator() -> None:
        # We assign variables to the function scope to avoid the overhead of
        # looking them up in the closure scope during the hot loop. We also
        # reuse the same list to avoid the overhead of allocating a new list
        # every time, and have it in shared memory to avoid the overhead of
        # copying the list to the main thread.
        _result_append = result.append
        _iterator_next = iterator.__next__
        _time_monotonic = time.monotonic
        _target_t = _time_monotonic() + target_dt

        try:
            while _time_monotonic() < _target_t:
                _result_append(_iterator_next())
        except StopIteration:
            raise StopAsyncIteration

    while True:
        try:
            # Run the iterator in a thread until we've reached the target time.
            await loop.run_in_executor(None, run_iterator)

            # Yield the results.
            for item in result:
                yield item

            # Clear the results.
            result.clear()
        except StopAsyncIteration:
            # Yield the remaining items in the result buffer.
            for item in result:
                yield item
            return


def ensure_async_iterator(iterable: Iterable[_T] | AsyncIterable[_T]) -> AsyncIterator[_T]:
    """Given an iterable or async iterable, return an async iterable.

    Args:
        iterable: The iterable to ensure is async.
        to_thread: Whether to run the iterable in a thread, if it is sync.

    Returns:
        An async iterable that runs the original iterable.
    """
    if isinstance(iterable, AsyncIterable):
        return aiter(iterable)

    return aiter(iter_to_aiter(iterable))


def merge_async_iterables(*async_iters: Iterable[_T] | AsyncIterable[_T]) -> AsyncIterator[_T]:
    """Merge multiple iterables or async iterables into one, yielding items as they are received.

    Args:
        async_iters: The async iterators to merge.

    Yields:
        Items from the async iterators, as they are received.

    Examples:
        >>> async def a():
        ...     for i in range(3):
        ...         await asyncio.sleep(0.07)
        ...         yield i
        >>> async def b():
        ...     for i in range(100, 106):
        ...         await asyncio.sleep(0.03)
        ...         yield i
        >>> async def demo_amerge():
        ...     async for item in merge_async_iterables(a(), b()):
        ...         print(item)
        >>> asyncio.run(demo_amerge())
        100
        101
        0
        102
        103
        1
        104
        105
        2
    """

    async def _inner() -> AsyncIterator[_T]:
        futs: dict[asyncio.Future[_T], AsyncIterator[_T]] = {}
        for it in async_iters:
            async_it = ensure_async_iterator(it)
            fut = asyncio.ensure_future(anext(async_it))
            futs[fut] = async_it

        while futs:
            done, _ = await asyncio.wait(futs, return_when=asyncio.FIRST_COMPLETED)
            for done_fut in done:
                try:
                    yield done_fut.result()
                except StopAsyncIteration:
                    pass
                else:
                    fut = asyncio.ensure_future(anext(futs[done_fut]))
                    futs[fut] = futs[done_fut]
                finally:
                    del futs[done_fut]

    return _inner()


if __name__ == "__main__":

    if not TYPE_CHECKING:

        def reveal_type(obj: object) -> None:
            print(f"{obj!r} is {type(obj)}")

    async def my_async_fn(arg1: _T, arg2: list[_T]) -> _T:
        for i in arg2:
            print(i)
        return arg1

    def my_sync_fn(arg1: _T, arg2: list[_T]) -> _T:
        for i in arg2:
            print(i)
        return arg1

    my_async_ensured = ensure_coroutine_function(my_async_fn)
    my_sync_ensured = ensure_coroutine_function(my_sync_fn, to_thread=False)
    my_sync_ensured_to_thread = ensure_coroutine_function(my_sync_fn, to_thread=True)

    # reveal_type(my_async_ensured)
    # reveal_type(my_sync_ensured)
    # reveal_type(my_sync_ensured_to_thread)

    async def main() -> None:

        a = await my_async_ensured(4, [1, 3, 2])
        b = await my_sync_ensured(4, [1, 3, 2])
        c = await my_sync_ensured_to_thread(4, [1, 3, 2])

        print(a, b, c)

        # reveal_type(a)  # asynkets/utils.py:94: note: Revealed type is "builtins.int"
        # reveal_type(b)  # asynkets/utils.py:95: note: Revealed type is "builtins.int"
        # reveal_type(c)  # asynkets/utils.py:96: note: Revealed type is "builtins.int"

    asyncio.run(main())

__all__ = ("ensure_coroutine_function",)



from asyncio import Event
from functools import partialmethod


class Switch:
    """A switch that can be turned on or off, and allows waiting for a state change.

    The switch can be in one of two states: on or off. It can be set to a specific state using the
    set_state() method, or turned on using the set() method and turned off using the clear()
    method. The current state of the switch can be checked using the is_set() and is_clear() methods.

    The switch also provides async methods for waiting until it is in a specific state. The wait() and
    wait_clear() methods can be used to wait until the switch is turned on or off, respectively. The
    wait_for() method can be used to wait for the switch to be in a specific state, and the
    wait_toggle(), wait_toggle_to(), wait_toggled_on(), and wait_toggled_off() methods can be
    used to wait for the switch to change from one state to another.
    """

    def __init__(self, initial_state: bool = False) -> None:
        super().__init__()
        self._ev_on = Event()
        self._ev_off = Event()
        self._state: bool = initial_state
        self.set_state(initial_state)

    def set_state(self, state: bool) -> None:
        """Set the switch to a specific state.

        Args:
            state: The state to set.
        """
        if state:
            self.set()
        else:
            self.clear()

    def set(self) -> None:
        """Turn the switch on."""

        if not self._state:
            self._state = True
            self._ev_on.set()
            self._ev_off.clear()

    def clear(self) -> None:
        """Turn the switch off."""
        if self._state:
            self._state = False
            self._ev_off.set()
            self._ev_on.clear()

    def is_set(self) -> bool:
        """Return True if the switch is on."""
        return self._state

    def is_clear(self) -> bool:
        """Return True if the switch is off."""
        return not self._state

    async def wait(self) -> None:
        """Wait until the switch is in the "on" state.

        Returns immediately if the switch is already in the "on" state.
        """
        if not self._state:
            await self._ev_on.wait()

    async def wait_clear(self) -> None:
        """Wait until the switch is in the "off" state.

        Returns immediately if the switch is already in the "off" state.
        """
        if self._state:
            await self._ev_off.wait()

    async def wait_for(self, state: bool) -> None:
        """Wait until the switch is in a specific state. Returns immediately if it already is.

        Args:
            state: The state to wait for.
        """
        if state:
            await self.wait()
        else:
            await self.wait_clear()

    async def wait_toggle(self) -> None:
        """Wait for the switch to change from False to True, or True to False."""
        if self._state:
            await self.wait_clear()
        else:
            await self.wait()

    async def wait_toggle_to(self, state: bool) -> None:
        """Wait for the switch to change from one state to another.

        Does not return immediately if the switch is already in the target state, but waits for the
        next change.

        Args:
            state: The state to wait for.
        """
        if state:
            await self.wait_clear()
            await self.wait()
        else:
            await self.wait()
            await self.wait_clear()

    wait_toggled_on = partialmethod(wait_toggle_to, True)
    wait_toggled_off = partialmethod(wait_toggle_to, False)


__all__ = ("Switch",)



from __future__ import annotations

from typing import cast, SupportsInt

from .switch import Switch


class EventfulCounter(SupportsInt):
    """EventfulCounter

    This class represents a counter that can be incremented or decremented, and
    also has (optional) minimum and maximum values that it can reach.

    The counter has SwitchEvent instances for being at/beyond the minimum and
    maximum values. This allows waiting for the counter to reach the minimum value,
    to reach the maximum value, to cross from below the minimum value to above the
    minimum value, or to cross from above to below the maximum value.

    The counter can be incremented or decremented by any amount with the inc() and
    dec() methods, or by using the += and -= operators. The counter can also be
    set to a specific value with the set() method.

    If the clamp_to_bounds flag is set, the counter will be clamped to the minimum
    and maximum values when it reaches them. Otherwise, it will continue to
    increment or decrement past them.

    This class inherits from SupportsFloat, which means it can be cast to a
    int using the int() function. It also supports comparison using the
    `>`, `<`, and `==` operators. It can be cast to a boolean using the
    bool() function, which returns True if the counter has a value greater
    than 0.

    Args:
        initial_value: The initial value of the counter. Defaults to 0.

        max_value: The maximum value that the counter can reach, or None. If none, the
            counter has no maximum value. Defaults to None.

        min_value: The minimum value that the counter can reach, or None. If none, the
            counter has no minimum value. Defaults to None.

        clamp_to_bounds: Whether the counter should be clamped to the minimum and maximum values.
            Defaults to False.
    """

    def __init__(
        self,
        initial_value: int = 0,
        max_value: int | None = None,
        min_value: int | None = None,
        clamp_to_bounds: bool = False,
    ) -> None:
        self._counter: int = initial_value
        self._max_value = max_value
        self._min_value = min_value
        self._clamp_to_bounds = clamp_to_bounds

        if max_value is not None and min_value is not None and max_value < min_value:
            raise ValueError("max_value must be greater than min_value")

        if max_value is not None and initial_value > max_value and clamp_to_bounds:
            raise ValueError(
                "initial_value must be less than max_value when clamp_to_bounds is True"
            )

        if min_value is not None and initial_value < min_value and clamp_to_bounds:
            raise ValueError(
                "initial_value must be greater than min_value when clamp_to_bounds is True"
            )

        if min_value is not None and max_value is not None and min_value == max_value:
            raise ValueError("min_value and max_value cannot be equal")

        self._min_ev = Switch()
        self._max_ev = Switch()
        self.set(initial_value)

    def inc(self, by: SupportsInt = 1) -> None:
        """Increment the counter."""
        self.set(self._counter + int(by))

    def dec(self, by: SupportsInt = 1) -> None:
        """Decrement the counter."""
        self.set(self._counter - int(by))

    @property
    def min_value(self) -> int | None:
        """Return the minimum value of the counter."""
        return self._min_value

    @min_value.setter
    def min_value(self, value: int | None) -> None:
        if value is None:
            self._min_value = None
            self._min_ev.set_state(False)
        else:
            if (
                self._max_value is not None
                and value > self._max_value
                and self._clamp_to_bounds
            ):
                raise ValueError("min_value cannot be greater than max_value")

            if value > self._counter and self._clamp_to_bounds:
                raise ValueError(
                    "min_value cannot be greater than current value when clamp_to_bounds is True"
                )

            self._min_value = value
            self.set(self._counter)

    @property
    def max_value(self) -> int | None:
        """Return the maximum value of the counter."""
        return self._max_value

    @max_value.setter
    def max_value(self, value: int | None) -> None:
        if value is None:
            self._max_value = None
            self._max_ev.set_state(False)
        else:
            if (
                self._min_value is not None
                and value < self._min_value
                and self._clamp_to_bounds
            ):
                raise ValueError("max_value cannot be less than min_value")

            if value < self._counter and self._clamp_to_bounds:
                raise ValueError(
                    "max_value cannot be less than current value when clamp_to_bounds is True"
                )

            self._max_value = value
            self.set(self._counter)

    def set(self, value: int) -> None:
        """Set the counter to a specific value."""
        self._counter = value

        if self._max_value is not None and self._counter >= self._max_value:
            self._max_ev.set()
            if self._clamp_to_bounds:
                self._counter = self._max_value

        if self._min_ev.is_set() and self._counter > cast(int, self._min_value):
            self._min_ev.clear()

        if self._min_value is not None and self._counter <= self._min_value:
            self._min_ev.set()
            if self._clamp_to_bounds:
                self._counter = self._min_value

        if self._max_ev.is_set() and self._counter < cast(int, self._max_value):
            self._max_ev.clear()

    def __int__(self) -> int:
        return self._counter

    def __iadd__(self, other: SupportsInt) -> EventfulCounter:
        self.inc(by=other)
        return self

    def __isub__(self, other: SupportsInt) -> EventfulCounter:
        self.dec(by=other)
        return self

    def __gt__(self, other: SupportsInt) -> bool:
        return int(self) > int(other)

    def __lt__(self, other: SupportsInt) -> bool:
        return int(self) < int(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SupportsInt):
            return int(self) == int(other)
        return False

    def __bool__(self) -> bool:
        return bool(self._counter)

    def min_is_set(self) -> bool:
        """Return True if the counter is at or below the minimum value."""
        return self._min_ev.is_set()

    def max_is_set(self) -> bool:
        """Return True if the counter is at or above the maximum value."""
        return self._max_ev.is_set()

    async def wait_max(self) -> None:
        """Wait until the counter reaches its maximum value.

        Returns immediately if it already has.
        """
        await self._max_ev.wait()

    async def wait_min(self) -> None:
        """Wait until the counter reaches its minimum value.

        Returns immediately if it already has.
        """
        await self._min_ev.wait()

    async def wait_max_clear(self) -> None:
        """Wait for the counter to be below its maximum value.

        Returns immediately if it already is.
        """
        await self._max_ev.wait_clear()

    async def wait_min_clear(self) -> None:
        """Wait for the counter to be above its minimum value.

        Returns immediately if it already is.
        """
        await self._min_ev.wait_clear()

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__}: {self._counter} "
            f"(min: {self._min_value}, max: {self._max_value})>"
        )

    __repr__ = __str__


__all__ = ("EventfulCounter",)



from .utils import *
from .fuse import *
from .pulse import *
from .switch import *
from .eventful_counter import *
from .async_std import *

from . import utils
from . import fuse
from . import pulse
from . import switch
from . import eventful_counter
from . import async_std


__all__ = (  # pyright: ignore
    *eventful_counter.__all__,
    *fuse.__all__,
    *pulse.__all__,
    *switch.__all__,
    *utils.__all__,
    *async_std.__all__,
)



import asyncio
import collections


class Fuse:
    """Similar to asyncio.Event, but can only be set once."""

    def __init__(self) -> None:
        self._waiters: collections.deque[asyncio.Future[bool]] = collections.deque()
        self._value = False

    def set(self) -> None:
        """Set the fuse."""
        if not self._value:
            self._value = True

            for fut in self._waiters:
                if not fut.done():
                    fut.set_result(True)

    def is_set(self) -> bool:
        """Return True if the fuse is set."""
        return self._value

    async def wait(self) -> None:
        """Wait for the fuse to be set."""
        if self._value:
            return

        fut = asyncio.get_event_loop().create_future()
        self._waiters.append(fut)
        try:
            await fut
        finally:
            self._waiters.remove(fut)


__all__ = ("Fuse",)



from __future__ import annotations


class NoValueT:
    __slots__ = ()

    instance: NoValueT

    def __new__(cls) -> NoValueT:
        if not hasattr(cls, "instance"):
            cls.instance = super().__new__(cls)
        return cls.instance


NoValue = NoValueT()


__all__ = ("NoValue", "NoValueT")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asynkets']

package_data = \
{'': ['*']}

install_requires = \
['setuptools>=65.6.3,<66.0.0', 'wheel>=0.38.4,<0.39.0']

setup_kwargs = {
    'name': 'asynkets',
    'version': '0.1.2',
    'description': '',
    'long_description': "\n# asynkets\n\n`asynkets` is a Python library which provides miscellaneous utilities that are useful when writing `asyncio` code.\n\nProvided utilities include:\n\n- *`Switch`*: A class that allows waiting for state changes of a switch that can be turned on or off. It provides async methods for waiting until it is in a specific state, and for waiting until it switches to a different state.\n- *`EventfulCounter`*: A counter that can be incremented or decremented and has optional minimum and maximum values. It also provides async methods for waiting until some threshold (either low or high) is reached, or until some threshold is reached or left, by leveraging Switches.\n- *`Fuse`*: Similar to asyncio's Event, but can only be set once.\n- *`Pulse`*: A pulse that can be triggered and waited for. It can also be given a function to call when it is triggered and can be used as an async iterator.\n- *`PeriodicPulse`*: A pulse that fires periodically at a specified interval and delay. It can be closed and used as an async iterator like the Pulse class.\n- *`ensure_coroutine_function`* and *`ensure_async_iterator`*: return guaranteed async versions of a sync or async function or iterable, respectively. If the function or iterable is synchronous, it will be wrapped to be async; if it is already async, it will be returned as-is.\n- *`merge_async_iterables`*: Merges multiple iterables or async iterables into one, yielding items as they are received.\n\n\n## Installation\n\n`asynkets` is available on PyPI and can be installed with `pip` or your favorite package manager.\n\n```bash\npip install asynkets\n```\n\n## Usage\n\n(todo)\n\n## License\n\nThe library is provided under the MIT license. See the [LICENSE](LICENSE) file for more information.\n",
    'author': 'Pedro Batista',
    'author_email': 'pedrovhb@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Complete the class below with the missing collection methods.

from __future__ import annotations

import asyncio
import textwrap
from asyncio import Task
from collections import deque
from collections.abc import Sequence
from datetime import timedelta
from itertools import islice, pairwise
from typing import (
    AsyncIterable,
    Callable,
    Coroutine,
    Generator,
    Generic,
    Iterable,
    overload,
    SupportsIndex,
    TypeVar,
)

from .switch import Switch

_T = TypeVar("_T")

_SyncFn = Callable[[_T], None]
_AsyncFn = Callable[[_T], Coroutine[object, object, None]]


class TimedDeque(Generic[_T], deque[_T]):
    """A TimedDeque works like a deque, but any appended items are removed after a
    certain period of time. The period is specified in seconds or as a timedelta.

    Callbacks can be added to the TimedDeque, which are called whenever an item is
    removed from the TimedDeque. The item that was removed is passed to the callback as
    an argument. The callback will also be called when the item is removed by other means,
    such as when the TimedDeque is cleared, when the item is popped, or when the item is
    removed by the deque being full (per maxlen) when a new item is appended. If the callback
    is a coroutine function, it will be scheduled as a task; otherwise, it will be scheduled
    via the event loop's call_soon method.

    The TimedDeque can be awaited, in which case it will block until the deque is empty.

    Notes:
        - Some deque methods have return types that are incompatible with those of its
            MutableSequence ABC (e.g. pop and __setitem__). These methods have a
            type: ignore[override] comment to suppress the mypy error.

    """

    def __init__(
        self,
        period: float | timedelta,
        iterable: Iterable[_T] = (),
        maxlen: int | None = None,
        callbacks: list[_SyncFn[_T] | _AsyncFn[_T]] | None = None,
    ) -> None:
        super().__init__(maxlen=maxlen)
        self._loop = asyncio.get_event_loop()

        self._period = period.total_seconds() if isinstance(period, timedelta) else period
        self._callbacks: list[_SyncFn[_T] | _AsyncFn[_T]] = callbacks or []
        self._pending_futs: dict[int, asyncio.Future[_T]] = {}

        self._empty = Switch(initial_state=len(self) == 0)

        self.extend(iterable)

    def __await__(self) -> Generator[None, None, None]:
        yield from self._empty.wait().__await__()

    def add_callback(self, callback: _SyncFn[_T] | _AsyncFn[_T]) -> None:
        self._callbacks.append(callback)

    def _resolve(self, fut: asyncio.Future[_T]) -> None:
        """Called when an item is removed. Schedules callbacks and pops the item's future."""
        item = fut.result()
        for callback in self._callbacks:
            if asyncio.iscoroutinefunction(callback):
                self._loop.create_task(callback(item))
            else:
                self._loop.call_soon(callback, item)

        if len(self) == 0:
            self._empty.set()

    def _remove_item(self, item: _T, remove_from_self: bool = False) -> None:
        """Remove an item from the deque and set its future's result.

        Args:
            item: The item to remove.
            remove_from_self: Whether to remove the item from the deque. This is necessary
                when the item is removed by the period expiring; in other cases, the item
                is already removed from the deque (e.g. when the item is popped).
        """

        if remove_from_self:
            super().remove(item)

        if id(item) in self._pending_futs:
            fut = self._pending_futs.pop(id(item))
            fut.set_result(item)

    def append(self, item: _T) -> None:

        if len(self) == self.maxlen:
            self._remove_item(self.popleft())

        super().append(item)
        self._empty.clear()

        new_fut = self._loop.create_future()
        self._pending_futs[id(item)] = new_fut
        new_fut.add_done_callback(self._resolve)
        self._loop.call_later(self._period, self._remove_item, item, True)

    def appendleft(self, item: _T) -> None:
        if len(self) == self.maxlen:
            self._remove_item(self.pop())

        super().appendleft(item)
        self._empty.clear()

        new_fut = self._loop.create_future()
        self._pending_futs[id(item)] = new_fut
        new_fut.add_done_callback(self._resolve)
        self._loop.call_later(self._period, self._remove_item, item, True)

    def clear(self) -> None:
        for item in self:
            self._remove_item(item)
        super().clear()
        self._empty.set()

    def pop(self) -> _T:  # type: ignore[override]
        item = super().pop()
        self._remove_item(item)
        return item

    def popleft(self) -> _T:
        item = super().popleft()
        self._remove_item(item)
        return item

    def remove(self, item: _T) -> None:
        super().remove(item)
        self._remove_item(item)

    def extend(self, iterable: Iterable[_T]) -> None:
        for item in iterable:
            self.append(item)

    def extendleft(self, iterable: Iterable[_T]) -> None:
        for item in iterable:
            self.appendleft(item)

    def __delitem__(self, index: SupportsIndex) -> None:  # type: ignore[override]
        item = self[index]
        super().__delitem__(index)
        self._remove_item(item)

    def __setitem__(self, index: SupportsIndex, value: _T) -> None:  # type: ignore[override]
        item = self[index]
        super().__setitem__(index, value)
        self._remove_item(item)

    def fill_from_async_iterable(self, iterable: AsyncIterable[_T]) -> Task[None]:
        """Fill the TimedDeque from an async iterable.

        Args:
            iterable: The async iterable to fill the TimedDeque from.
        """
        return self._loop.create_task(self._fill_from_async_iterable(iterable))

    async def _fill_from_async_iterable(self, iterable: AsyncIterable[_T]) -> None:
        async for item in iterable:
            self.append(item)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._period}, {super().__repr__()})"

    __str__ = __repr__


class TimeBucket(Generic[_T], Sequence[_T]):
    """A TimeBucket object is a list of TimedDeques. When an item is appended to a
    TimeBucket, it is appended to the first TimedDeque in the list. As it's removed from
    the TimedDeque, it is appended to the next TimedDeque in the list. This continues until
    the item is removed from the last TimedDeque.
    """

    def __init__(
        self,
        period: float | timedelta,
        num_buckets: int,
        iterable: Iterable[_T] = (),
    ) -> None:
        self._period = period.total_seconds() if isinstance(period, timedelta) else period
        self._empty = Switch(initial_state=True)

        self._loop = asyncio.get_event_loop()
        self._buckets: tuple[TimedDeque[_T], ...] = tuple(
            TimedDeque(period=self._period) for _ in range(num_buckets)
        )

        for a, b in pairwise(self._buckets):
            a.add_callback(b.appendleft)
        self._buckets[-1].add_callback(lambda _: self._empty.set() if len(self) == 0 else None)

        for item in iterable:
            self.append(item)

    def __await__(self) -> Generator[None, None, None]:
        yield from self._empty.wait().__await__()

    def __len__(self) -> int:
        return sum(len(bucket) for bucket in self._buckets)

    @overload
    def __getitem__(self, index: SupportsIndex | int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[_T]:
        ...

    def __getitem__(self, index: SupportsIndex | int | slice) -> _T | list[_T]:
        if isinstance(index, slice):
            return [item for bucket in self._buckets[index] for item in bucket]
        else:
            int_index = int(index)

            # todo - optimize this to skip buckets as in the commented code below
            if int_index >= 0:
                return next(islice(self, int_index, int_index + 1))
            else:
                return next(islice(reversed(self), -int_index, -int_index + 1))

            # for bucket in self._buckets:
            #     if int_index < len(bucket):
            #         return bucket[int_index]
            #     else:
            #         int_index -= len(bucket)
            # raise IndexError("index out of range")

    def __iter__(self) -> Generator[_T, None, None]:
        for bucket in reversed(self._buckets):
            yield from reversed(bucket)

    def __reversed__(self) -> Generator[_T, None, None]:
        for bucket in self._buckets:
            yield from bucket

    def append(self, item: _T) -> None:
        self._buckets[0].appendleft(item)
        self._empty.clear()

    def fill_from_async_iterable(self, iterable: AsyncIterable[_T]) -> Task[None]:
        """Fill the TimedDeque from an async iterable.

        Args:
            iterable: The async iterable to fill the TimedDeque from.
        """
        return self._loop.create_task(self._fill_from_async_iterable(iterable))

    async def _fill_from_async_iterable(self, iterable: AsyncIterable[_T]) -> None:
        async for item in iterable:
            self.append(item)

    def __repr__(self) -> str:
        bucket_strs: list[str] = []
        for i, bucket in enumerate(self._buckets):
            bucket_strs.append(f"T - {i * self._period}: {list(bucket)}")
            # todo - handle 0.30000000000000004
        s = textwrap.indent("\n".join(bucket_strs), " " * 4)
        return (
            f"{self.__class__.__name__}"
            f"({self._period}s window, {len(self._buckets)} buckets, {len(self)} items) [\n{s}\n]"
        )

    __str__ = __repr__


if __name__ == "__main__":

    async def main() -> None:
        td = TimedDeque(1, range(10))

        async def filler() -> None:
            for i in range(50):
                td.append(i)
                await asyncio.sleep(0.01)
                print(f"Appended {i}")

            print("Done appending")

        asyncio.create_task(filler())

        td.add_callback(print)

        await td
        print("deque is empty")

        time_bucket = TimeBucket(0.1, 5, range(10))

        for i in range(50):
            time_bucket.append(i)
            await asyncio.sleep(0.01)
            print(time_bucket)

        for item in time_bucket:
            print(item)
        print(
            time_bucket[0],
            time_bucket[1],
            time_bucket[2],
            time_bucket[3],
            time_bucket[4],
            time_bucket[20],
            time_bucket[-2],
        )

    asyncio.run(main())


__all__ = ("TimeBucket", "TimedDeque")



from __future__ import annotations

import time
import asyncio
from asyncio import Future
from collections import deque
from datetime import timedelta
from functools import partial
from typing import AsyncIterator, Callable, cast, Coroutine, Generator

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

from .fuse import Fuse

_P = ParamSpec("_P")


class PulseClosed(Exception):
    """Raised when a pulse is closed."""


class _BasePulse:
    def __init__(self) -> None:
        self._waiters: deque[Future[float]] = deque()
        self._value: float | None = None
        self._loop = asyncio.get_event_loop()
        self._closed = Fuse()
        self._pulse_callbacks: list[Callable[[float], object] | Callable[[], object]] = []

    def add_pulse_callback(
        self,
        callback: Callable[_P, object],
        thread_safe: bool = False,
        *__args: _P.args,
        **__kwargs: _P.kwargs,
    ) -> None:
        """Add a callback to be called when the pulse is fired.

        If it is a coroutine, it will be scheduled as a task; otherwise, it will be scheduled
        via the event loop's call_soon method if thread_safe is False, or via call_soon_threadsafe
        if thread_safe is True.
        """

        if asyncio.iscoroutinefunction(callback):
            _fn = cast(Callable[_P, Coroutine[object, object, object]], callback)

            def _cb() -> None:
                _ = asyncio.create_task(_fn(*__args, **__kwargs))

        elif thread_safe:

            def _cb() -> None:
                self._loop.call_soon_threadsafe(partial(callback, *__args, **__kwargs))

        else:

            def _cb() -> None:
                self._loop.call_soon(partial(callback, *__args, **__kwargs))

        self._pulse_callbacks.append(_cb)

    def _run_callbacks(self) -> None:
        for cb in self._pulse_callbacks:
            self._loop.call_soon(cb)

    @property
    def is_closed(self) -> bool:
        """Return True if the pulse is closed."""
        return self._closed.is_set()

    def close(self) -> None:
        """Close the pulse, waking up all waiters.

        The resulting future will have a PulseClosed exception set as its exception.
        """
        self._closed.set()
        for fut in self._waiters:
            fut.set_exception(PulseClosed())

    async def wait_closed(self) -> None:
        """Wait for the pulse to be closed."""
        await self._closed.wait()

    def wait(self) -> Future[float]:
        """Wait for the pulse to be fired.

        Returns a future that will be resolved when the pulse is fired.
        The future's result will be the time at which the pulse was fired, as per time.time().
        """
        fut = asyncio.get_event_loop().create_future()
        self._waiters.append(fut)
        return fut

    def __await__(self) -> Generator[None, None, float]:
        """Wait for the pulse to be fired.

        Returns the time at which the pulse was fired, as given by time.time().
        """
        return self.wait().__await__()

    async def _aiter(self) -> AsyncIterator[float]:
        while True:
            try:
                yield await self
            except PulseClosed:
                break

    def __aiter__(self) -> AsyncIterator[float]:
        return self._aiter()

    def _fire(self) -> None:
        """Fire the pulse, waking up all waiters.

        The resulting future will be resolved with the current time.
        """
        if self._closed.is_set():
            raise RuntimeError("Cannot fire a closed pulse")

        self._value = time.time()
        for fut in self._waiters:
            fut.set_result(self._value)
        self._waiters.clear()

        self._run_callbacks()


class Pulse(_BasePulse):
    """A pulse that can be triggered and waited for.

    Waiting for a pulse will block until the pulse is triggered, and will return
    the time at which the pulse was triggered. Alternatively, the pulse can be given
    a function to call when it is triggered. In this case, the return value of waiting
    on the pulse will be the result of calling the function.

    The pulse can be closed, which will cause all waiters to be woken up with a
    PulseClosed exception. After the pulse is closed, it cannot be fired again.

    The pulse can be used as an async iterator, which will yield the time at which
    the pulse is fired. The iterator will stop yielding when the pulse is closed.

    Examples:
        >>> pulse = Pulse()
        >>> pulse.fire()
        >>> await pulse
        123.456

        >>> pulse = Pulse()
        >>> pulse.add_pulse_callback(lambda: print("Pulse fired!"))
        >>> pulse.fire()
        Pulse fired!

        >>> pulse = Pulse()
        >>> async def pulse_subscriber(pulse: Pulse) -> None:
        ...     async for t in pulse:
        ...         print(t)
        >>> asyncio.create_task(pulse_subscriber(pulse))
        >>> pulse.fire()
        123.456
        >>> pulse.fire()
        123.457
        >>> pulse.close()
        >>> pulse.fire()
        Traceback (most recent call last):
        ...
        RuntimeError: Cannot fire a closed pulse
    """

    fire = _BasePulse._fire


class PeriodicPulse(_BasePulse):
    """A pulse that fires periodically.

    The pulse will fire every `interval` seconds, and will wait for `delay` seconds
    before the first pulse is fired. This can be used to implement a periodic task,
    and is different from asyncio's `loop.call_later` or `while True: await asyncio.sleep(x)`
    in that it will not drift over time.

    The pulse can be closed, which will cause all waiters to be woken up with a
    PulseClosed exception. After the pulse is closed, it does not fire again.

    The pulse can be used as an async iterator, which will yield the time at which
    the pulse is fired. The iterator will stop yielding when the pulse is closed.

    Examples:
        >>> pulse = PeriodicPulse(1.0)
        >>> async for t in pulse:
        ...     print(t)
        123.456
        124.456
        125.456

        >>> pulse = PeriodicPulse(period=timedelta(minutes=5), start_delay=0.5)
        >>> pulse.add_pulse_callback(lambda: print(f"Pulse fired! {time.time()}"))
        >>> await asyncio.sleep(60 * 15)
        Pulse fired! 12:00:00
        Pulse fired! 12:05:00
        Pulse fired! 12:10:00
    """

    def __init__(
        self,
        period: float | timedelta,
        start_delay: float | timedelta | None = None,
    ) -> None:
        """Create a periodic pulse.

        Args:
            period: The period of the pulse, in seconds or a timedelta.
            start_delay: The delay before the first pulse is fired. If None, the first pulse
                will be fired after `period` seconds. To fire the first pulse immediately,
                pass 0.
        """
        super().__init__()
        self._period = period.total_seconds() if isinstance(period, timedelta) else period

        self._ticks = 0

        if start_delay is None:
            self._start_delay = self._period
        elif isinstance(start_delay, timedelta):
            self._start_delay = start_delay.total_seconds()
        else:
            self._start_delay = start_delay

        self._start_time = self._loop.time()
        self._next_tick_handle = self._loop.call_at(
            self._start_time + self._start_delay,
            self._tick,
        )
        self._target_period: float | None = None

    def _tick(self) -> None:
        if self._closed.is_set():
            return

        self._fire()

        if self._target_period is not None:
            self._period = self._target_period
            self._target_period = None
            self._ticks = 0
            self._start_time = self._loop.time()
            self._start_delay = 0

        self._ticks += 1
        self._next_tick_handle = self._loop.call_at(
            when=self._start_time + self._period * self._ticks + self._start_delay,
            callback=self._tick,
        )

    @property
    def period(self) -> timedelta:
        """The period of the pulse."""
        return timedelta(seconds=self._period)

    @period.setter
    def period(self, period: float | timedelta) -> None:
        self._target_period = period if isinstance(period, float) else period.total_seconds()


if __name__ == "__main__":
    import time

    async def main() -> None:
        pulse = Pulse()

        async def show_pulses() -> None:
            async for t in pulse:
                print(f"pulse1 fired at {t}")
            print("pulse1 finished")

        for _ in range(3):
            asyncio.create_task(show_pulses())

        pulse.fire()
        await asyncio.sleep(0.1)
        pulse.fire()
        await asyncio.sleep(0.1)
        pulse.fire()
        await asyncio.sleep(0.3)
        pulse.fire()
        await asyncio.sleep(0.3)
        pulse.close()

        time_pulse = PeriodicPulse(0.25)

        async def show_time_pulses() -> None:
            # async for t in time_pulse:
            # print(f"time_pulse fired at {t}")
            t = await time_pulse
            print(f"time_pulse fired at {t}")
            print("time_pulse finished")

        for _ in range(3):
            asyncio.create_task(show_time_pulses())

        await asyncio.sleep(3)

        print("awaiting once", time.time())
        await time_pulse
        print("awaiting twice", time.time())
        await time_pulse
        print("awaiting thrice", time.time())
        await time_pulse
        print("awaiting finished", time.time())
        time_pulse.close()

        async def main_test_set_period() -> None:
            t = time.time()
            start_delay = 1 - divmod(t, 1)[1]  # start at the next second, for round numbers
            # await asyncio.sleep(start_delay)
            pulse = PeriodicPulse(0.1, start_delay=start_delay)
            start_time = time.time()
            print(f"start_time: {start_time}; period: {pulse.period}")
            pulse.add_pulse_callback(
                lambda: print(f"pulse fired at t+{time.time()-start_time:3.5f}")
            )
            await asyncio.sleep(1.3)
            print("setting period to 0.5")
            pulse.period = timedelta(seconds=0.5)
            await asyncio.sleep(5)
            pulse.close()

        await main_test_set_period()

    asyncio.run(main())


__all__ = ("PeriodicPulse", "Pulse", "PulseClosed")
